/*********************************************************************
 *  HTTP Header File for Microchip TCP/IP Stack HTTP2 Server
 *
 * Author: Craig A. Lindley
 * Last Update: 11/18/2007
 ********************************************************************/

#ifndef __HTTP2_H
#define __HTTP2_H

#include "TCPIP Stack/TCPIP.h"

/*********************************************************************
 * Server Configuration Settings
 ********************************************************************/
	#define HTTP_PORT               (80u)	// Listening port for HTTP server
	#define HTTP_MAX_DATA_LEN		(100u)	// Max bytes to store get and cookie args
	#define HTTP_MAX_FN_LEN			(64u)	// Max path/filename length
	#define HTTP_CACHE_LEN			("600")	// Max lifetime (sec) of static responses as string
	#define HTTP_TIMEOUT			(45u)	// Max time (sec) to await more data before

	// Authentication requires Base64 decoding
	#define STACK_USE_BASE64_DECODE

/*********************************************************************
 * Commands and Server Responses
 ********************************************************************/
	// Supported Commands and Server Response Codes
	typedef enum _HTTP_STATUS
	{
	    HTTP_GET = 0u,					// GET command is being processed
	    HTTP_POST,						// POST command is being processed
		HTTP_UNAUTHORIZED,				// 401 Unauthorized will be returned
	    HTTP_NOT_FOUND,					// 404 Not Found will be returned
		HTTP_OVERFLOW,					// 414 Request-URI Too Long will be returned
		HTTP_INTERNAL_SERVER_ERROR,		// 500 Internal Server Error will be returned
		HTTP_NOT_IMPLEMENTED,			// 501 Not Implemented (not a GET or POST command)
		HTTP_REDIRECT,					// 302 Redirect will be returned
		HTTP_SSL_REQUIRED				// 403 Forbidden is returned, indicating SSL is required
	} HTTP_STATUS;
	
/*********************************************************************
 * HTTP State Definitions
 ********************************************************************/

	// Basic HTTP Connection State Machine
	typedef enum _SM_HTTP2
	{
		SM_HTTP_IDLE = 0u,				// Socket is idle
		SM_HTTP_PARSE_REQUEST,			// Parses the first line for a file name and GET args
		SM_HTTP_PARSE_HEADERS,			// Reads and parses headers one at a time
		SM_HTTP_AUTHENTICATE,			// Validates the current authorization state
		SM_HTTP_PROCESS_GET,			// Invokes user callback for GET args or cookies
		SM_HTTP_PROCESS_REQUEST,		// Begins the process of returning data
		SM_HTTP_SERVE_HEADERS,			// Sends any required headers for the response
		SM_HTTP_SERVE_BODY,				// Serves the actual content
		SM_HTTP_DISCONNECT,				// Disconnects the server and closes all files
		SM_HTTP_WAIT					// Unused state
	} SM_HTTP2;
	
	// Result States for Read and Write Operations
	typedef enum _HTTP_IO_RESULT
	{
		HTTP_IO_DONE = 0u,	// Finished with procedure
		HTTP_IO_NEED_DATA,	// More data needed to continue, call again later
		HTTP_IO_WAITING,	// Waiting for asynchronous process to complete, call again later
	} HTTP_IO_RESULT;

	// File Type Definitions
	typedef enum _HTTP_FILE_TYPE
	{
		HTTP_TXT = 0u,		// File is a text document
		HTTP_HTM,			// File is HTML (extension .htm)
		HTTP_HTML,			// File is HTML (extension .html)
		HTTP_CGI,			// File is HTML (extension .cgi)
		HTTP_XML,			// File is XML (extension .xml)
		HTTP_CSS,			// File is stylesheet (extension .css)
		HTTP_GIF,			// File is GIF image (extension .gif)
		HTTP_PNG,			// File is PNG image (extension .png)
		HTTP_JPG,			// File is JPG image (extension .jpg)
		HTTP_JAVA,			// File is java (extension .java)
		HTTP_WAV,			// File is audio (extension .wav)
		HTTP_UNKNOWN		// File type is unknown
	} HTTP_FILE_TYPE;

	// HTTP Connection Struct
	// Stores partial state data for each connection
	// Meant for storage in fast access RAM
	typedef struct _HTTP_STUB
	{
	    SM_HTTP2 sm;						// Current connection state
	    TCP_SOCKET socket;					// Socket being served
	} HTTP_STUB;

	#define sktHTTP		httpStubs[curHTTPID].socket		// Access the current socket
	#define smHTTP		httpStubs[curHTTPID].sm			// Access the current state machine

	// Stores extended state data for each connection
	typedef struct _HTTP_CONN
	{
		DWORD ticcount;						// Tic count used as connection watchdog timer
		BYTE *ptrData;						// Points to first free byte in data
		char fn[HTTP_MAX_FN_LEN];			// Storage for the requested file
		BOOL fileExists;					// True if requested file exists; false otherwise
		DWORD size;							// The size of the file
		BYTE handle;						// Handle to read file with
	    DWORD offset;						// Offset into file for next data read
		BYTE hasArgs;						// True if there were get or cookie arguments	
		BYTE isAuthorized;					// 0x00-0x79 on fail, 0x80-0xff on pass
		HTTP_STATUS httpStatus;				// Request method/status
	    HTTP_FILE_TYPE fileType;			// File type to return with Content-Type
		BYTE data[HTTP_MAX_DATA_LEN];		// General purpose data buffer
	} HTTP_CONN;
	
	#define RESERVED_HTTP_MEMORY ( (DWORD)MAX_HTTP_CONNECTIONS * (DWORD)sizeof(HTTP_CONN))

	// Function prototypes
	void HTTPInit(void);
	void HTTPServer(void);
#endif
