// FAT File System Interface Header File
// Filesystem based on USBwiz OEM board
// Written by: Craig Lindley
// Last Update: 11/12/2007

#include <string.h>
#include "GenericTypeDefs.h"
#include "TCPIP Stack/HardwareProfile.h"
#include "TCPIP Stack/USBwiz_lib.h"
#include "TCPIP Stack/Ffs.h"

#define MAX_FILE_HANDLES 4

// Create an array of booleans to indicate whether a
// file handle is available. A value of TRUE indicates
// the handle of this index is in use.
BOOL FILE_HANDLES[MAX_FILE_HANDLES];

// Initialize the USBwiz hardware in preparation for use
// Returns error code on error or no error on success
int8 _USBwiz_Initialize(void)
{
	int8 c;
	
	// Since MODE 0,1 pins are not connected
	// USBwiz is in SPI mode like we want it to be

	// Pull USBwiz reset pin low and release
	USBW_RST_IO = 0;
	GHI_Sleep(50);
	USBW_RST_IO = 1;
	GHI_Sleep(50);	
	
	// Close interface just in case!
	GHI_CloseInterface();
	
	// Lets get the interface going
	c = GHI_OpenInterface();
	if (c != FFS_NO_ERROR) {
		return c;
	}
	
	// This shouldn't be allowed to loop forever
	while(1) {
		c = GHI_GetC();
		if(c == '!')
			break;
	}
	
	// We need to get "00/r"
	c = 0;
	c =  (GHI_GetC() - '0') << 4;
	c |= (GHI_GetC() - '0');

	// Check for 00
	if(c != ERROR_NO_ERROR) {
		return FFS_HWD_INIT_ERROR;
	}

	// Now check for return char
	if(GHI_GetC() != '\r') {
		return FFS_HWD_INIT_ERROR;
	}
	
	// Indicate all file handles are available
	for (c = 0; c < MAX_FILE_HANDLES; c++)
		FILE_HANDLES[c] = FALSE;

	// All is well
	return FFS_NO_ERROR;
}

// Initialize the USBwiz hardware for operation
// and mount the specified device
BYTE FFS_Init(BYTE device) {

	int8 c;
	int16 version;

	// All analog pins to digital
	ADCON1=0x0f;
    
	// Shut off all comparators
	CMCON=0x07;
    
	// Configure I/O pins

	// Configure LED I/O pins
	LED0_TRIS = 0;  // make led pin output
	LED1_TRIS = 0;  // make led pin output

	// Configure USBwiz I/O pins
	USBW_RST_TRIS = 0; // make reset pin output
	USBW_CS_TRIS = 0;  // make chip select pin output
	USBW_DATARDY_TRIS = 1; // Make data ready pin an input
	USBW_BUSY_TRIS = 1; // Make busy pin an input

	// Set initial conditions

	// LEDs to off
	LED0_IO = 0;
	LED1_IO = 0;

	// USBwiz RST low; SSEL high
	USBW_RST_IO = 0;
	USBW_CS_IO = 1;

	// PICTail RST low; CS high
	ENC_RST_IO = 0;
	ENC_CS_IO = 1;

	// Initialize USBwiz
	c = _USBwiz_Initialize();
	if(c != FFS_NO_ERROR) {
		return c;
	}
	// Get library version just to make sure
	if (GHI_GetLibraryVersion() != GHI_USBWIZ_LIBRARY_VERSION) {
		return FFS_VERSION_ERROR;
	}
	// Attempt to mount USB port 1
	c = GHI_MountFATFileSystem(device);
	if (c != ERROR_NO_ERROR) {
		return c;
	}
	return FFS_NO_ERROR;
}

// The following are support functions

int8 _ChangeDirectory(int8 *path) {

	int8 c, len;
	int8 dir[12];
	int8 *ptr1, *ptr2;

	// Change dir to root
	strcpypgm2ram((char *) dir, (const far rom char *) "\\");
	c = GHI_ChangeDirectory(dir);

	// Check for error
	if(c != FFS_NO_ERROR) {
		return FFS_FNF_ERROR;
	}
	ptr1 = path;

	// If first path char is slash, move past it
	if (*ptr1 == '/') {
		ptr1++;
	}
	while (1) {
		// Is there another slash in the path ?
		ptr2 = (int8 *) strchr((char *) ptr1, '/');

		// If no slash we are done
		if (!ptr2) {
			return FFS_NO_ERROR;
		}
		// Found slash in the path get path segment name into dir
		len = ptr2 - ptr1;
		strncpy((char *) dir, (char *) ptr1, len);

		// Null terminate name
		*(dir + len) = '\0';

		// Attempt to change directory
		c = GHI_ChangeDirectory(dir);

		// Check for error
		if(c != FFS_NO_ERROR) {
			return FFS_FNF_ERROR;
		}
		// Advance ptrs
		ptr1 = ptr2 + 1;
	}
	return FFS_NO_ERROR;
}

int8 *_GetFilenameFromPath(int8 *path) {

	int8 *ptr;

	// Find the last slash in the path
	ptr = (int8 *) strrchr((char *) path, '/');

	return (ptr == 0) ? path : (ptr + 1);
}

// Determine if a file or directory exists
// name can include a path specification and filename
// Returns TRUE if the item was found; FALSE if not
BOOL FFS_FileExists(char *pathfile, DWORD *size) {

	int8 c, *ptr, attrs;
	int32 timedate;

	if (!pathfile) {
		return FFS_FNF_ERROR;
	}
	// Change to the dir containing the file
	c = _ChangeDirectory((int8 *) pathfile);
	if (c != FFS_NO_ERROR) {
		return FALSE;
	}
	ptr = _GetFilenameFromPath((int8 *) pathfile);

	c = GHI_GetFileInfo(ptr, size, &attrs, &timedate);

	return (c == ERROR_NO_ERROR) ? TRUE : FALSE;
}

// Attempt to assign a handle to a file and open it for reading
BYTE FFS_GetFileHandle(char *pathfile, BYTE *handleptr) {

	int8 c, *ptr;

	// Does this file already have a handle?
	if (*handleptr != FFS_NO_HANDLE) {
		return FFS_NO_ERROR;
	}
	// File is not yet open for read

	// Find a file handle for use
	*handleptr = FFS_NO_HANDLE;
	for (c = 0; c < MAX_FILE_HANDLES; c++) {
		if (!FILE_HANDLES[c]) {
			FILE_HANDLES[c] = TRUE;
			*handleptr = c;
			break;
		}
	}
	if (*handleptr == FFS_NO_HANDLE) {
		return FFS_NO_HANDLE;
	}
	// Change to the dir containing the file
	c = _ChangeDirectory((int8 *) pathfile);
	if (c != FFS_NO_ERROR) {
		return FFS_DNF_ERROR;
	}
	// Extract filename from pathfile
	ptr = _GetFilenameFromPath((int8 *) pathfile);

	// Attempt to open the file
	c = GHI_OpenFile(*handleptr, ptr, FILE_READ_MODE);
	if (c != ERROR_NO_ERROR) {
		return FFS_OPEN_ERROR;
	}
	return FFS_NO_ERROR;
}
	
// Read from the open file. This will never be called
// unless the handle is valid.
BYTE FFS_ReadFile(BYTE *handleptr, BYTE *buffer, DWORD *offset, DWORD reqsize, DWORD *readsize ) {

	int8 c;
	
	// Have we reached EOF already ?
	if (*offset == FFS_EOF) {
		// EOF reached do some cleanup
		*readsize = 0;
		FFS_CleanUp(handleptr);
		return FFS_EOF;
	}
	// Seek to position of last read
	if (*offset != 0) {
		c = GHI_SeekFile(*handleptr, *offset);
		if (c != ERROR_NO_ERROR) {
			c = FFS_SEEK_ERROR;
			goto error;
		}
	}
	// Now read the file data
	c = GHI_ReadFile(*handleptr, buffer, reqsize, '*', readsize);
	if (c != ERROR_NO_ERROR) {
		c = FFS_READ_ERROR;
		goto error;
	}
	// Do some cleanup
	if (*readsize != reqsize) {
		*offset = FFS_EOF;
	}	else	{
		*offset += *readsize;
	}
	return FFS_NO_ERROR;

	// An error occurred close up shop
error:
	FFS_CleanUp(handleptr);
	return c;
}

BYTE FFS_CleanUp(BYTE *handleptr) {

	int8 handle;

	handle = *handleptr;

	if (handle == FFS_NO_HANDLE) {
		return FFS_NO_ERROR;
	}
	// Do the cleanup
	GHI_CloseFile(handle);
	FILE_HANDLES[handle] = FALSE;
	*handleptr = FFS_NO_HANDLE;
	
	return FFS_NO_ERROR;
}