 /*********************************************************************
 * Modified version of Microchip HTTP2 Server Code
 * Able to serve up static pages via GET method
 * Filesystem based on USBwiz OEM board
 *
 * Author: Craig A. Lindley
 * Last Update: 11/26/2007
 ********************************************************************/


#include "TCPIP Stack/TCPIP.h"
#include "TCPIP Stack/Ffs.h"

/*********************************************************************
 * Constants
 ********************************************************************/

	static ROM BYTE HTTP_CRLF[] = "\r\n";	// New line sequence
	#define HTTP_CRLF_LEN	2				// Length of above string
	#define FILE_READ_SIZE 384

/*********************************************************************
 * File and Content Type Settings
 ********************************************************************/
	
	// Corresponding File Type Extensions
	static ROM char *httpFileExtensions[HTTP_UNKNOWN+1] =
	{
	    "txt",          // HTTP_TXT
	    "htm",          // HTTP_HTM
	    "html",         // HTTP_HTML
	    "cgi",          // HTTP_CGI
	    "xml",          // HTTP_XML
	    "css",          // HTTP_CSS
	    "gif",          // HTTP_GIF
	    "png",          // HTTP_PNG
	    "jpg",          // HTTP_JPG
	    "cla",          // HTTP_JAVA
	    "wav",          // HTTP_WAV
		"\0\0\0"		// HTTP_UNKNOWN
	};
	
	// Corresponding Content Type Strings
	static ROM char *httpContentTypes[HTTP_UNKNOWN+1] =
	{
	    "text/plain",            // HTTP_TXT
	    "text/html",             // HTTP_HTM
	    "text/html",             // HTTP_HTML
	    "text/html",             // HTTP_CGI
	    "text/xml",              // HTTP_XML
	    "text/css",              // HTTP_CSS
	    "image/gif",             // HTTP_GIF
	    "image/png",             // HTTP_PNG
	    "image/jpeg",            // HTTP_JPG
	    "application/java-vm",   // HTTP_JAVA
	    "audio/x-wave",          // HTTP_WAV
		""						 // HTTP_UNKNOWN
	};
		
/*********************************************************************
 * Commands and Server Responses
 ********************************************************************/

	// Corresponding initial response strings (to HTTP_STATUS enum)
	static ROM char *HTTPResponseHeaders[] =
	{
		"HTTP/1.1 200 OK\r\nConnection: close\r\n",
		"HTTP/1.1 200 OK\r\nConnection: close\r\n",
		"HTTP/1.1 401 Unauthorized\r\nWWW-Authenticate: Basic realm=\"Protected\"\r\n\r\n401 Unauthorized: Password required\r\n",
		"HTTP/1.1 404 Not found\r\n\r\n404: File not found\r\n",
		"HTTP/1.1 414 Request-URI Too Long\r\n\r\n414 Request-URI Too Long: Buffer overflow detected\r\n",
		"HTTP/1.1 500 Internal Server Error\r\n\r\n500 Internal Server Error: Expected data not present\r\n",
		"HTTP/1.1 501 Not Implemented\r\n\r\n501 Not Implemented: Only GET and POST supported\r\n",
		"HTTP/1.1 302 Found\r\nLocation: ",
		"HTTP/1.1 403 Forbidden\r\n\r\n403 Forbidden: SSL Required - use HTTPS\r\n"
	};
	
/*********************************************************************
 * Header Parsing Configuration
 ********************************************************************/
	#define HTTP_NUM_HEADERS		1
	
	// Header strings for which we'd like to parse
	static ROM char *HTTPRequestHeaders[HTTP_NUM_HEADERS] =
	{
		"Authorization:"
	};
	
	// Set to length of longest string above
	#define HTTP_MAX_HEADER_LEN		(14u)

/*********************************************************************
 * HTTP Connection State Global Variable
 ********************************************************************/
	#pragma udata HTTP_CONNECTION_STATES
	HTTP_CONN curHTTP;							// Current HTTP connection state
	HTTP_STUB httpStubs[MAX_HTTP_CONNECTIONS];	// HTTP stubs with state machine and socket
	BYTE curHTTPID;								// Which HTTP stub is active
	#pragma udata


/*********************************************************************
 * Data buffers - One for each of the file handles
 ********************************************************************/
	#pragma udata DATA_BUFFERS

	BYTE data0[FILE_READ_SIZE];
	BYTE data1[FILE_READ_SIZE];
	BYTE data2[FILE_READ_SIZE];
	BYTE data3[FILE_READ_SIZE];

	#pragma udata

/*********************************************************************
 * Function Prototypes
 ********************************************************************/

	// Internal function Prototypes
	static void HTTPHeaderParseLookup(BYTE i);
	static void HTTPHeaderParseAuthorization(void);
	static void HTTPProcess(void);
	static BOOL HTTPSendFile(void);
	static void HTTPLoadConn(BYTE connID);

	#define mMIN(a, b)	((a<b)?a:b)

/*********************************************************************
 * Function:        void HTTPInit(void)
 *
 * PreCondition:    TCP must already be initialized.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Sets all HTTP sockets to listening state.
 *                  Initialize state machine for each connection.
 *
 * Note:            This function is called only one during lifetime
 *                  of the application.
 ********************************************************************/
void HTTPInit(void)
{
	WORD oldPtr;

    for(curHTTPID = 0; curHTTPID < MAX_HTTP_CONNECTIONS; curHTTPID++) {

		smHTTP = SM_HTTP_IDLE;
		sktHTTP = TCPListen(HTTP_PORT);
		
	    // Save the default record (just invalid file handles)
    	oldPtr = MACSetWritePtr(BASE_HTTPB_ADDR + curHTTPID*sizeof(HTTP_CONN));
		MACPutArray((BYTE*)&curHTTP, sizeof(HTTP_CONN));
		MACSetWritePtr(oldPtr);
    }
}

/*********************************************************************
 * Function:        void HTTPServer(void)
 *
 * PreCondition:    HTTPInit() must already be called.
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Browse through each connection and try to bring
 *					it a step closer to completion.
 *
 * Note:            This function acts as a task (similar to one in
 *                  RTOS).  This function performs its task in
 *                  co-operative manner.  Main application must call
 *                  this function repeatdly to ensure all open
 *                  or new connections are served on time.
 ********************************************************************/
void HTTPServer(void)
{
	BYTE conn;

	for(conn = 0; conn < MAX_HTTP_CONNECTIONS; conn++)
	{
		// Determine if this connection is eligible for processing
		if(httpStubs[conn].sm != SM_HTTP_IDLE || TCPIsGetReady(httpStubs[conn].socket))
		{
			HTTPLoadConn(conn);
			HTTPProcess();
		}
	}
}

/*********************************************************************
 * Function:        static void HTTPLoadConn(BYTE connID)
 *
 * PreCondition:    None
 *
 * Input:           connID the connection ID to load
 *
 * Output:          curHTTP has a new connection loaded
 *
 * Side Effects:    None
 *
 * Overview:        Loads the current HTTP connection out of Ethernet
 *					buffer RAM and into local RAM for processing.
 *
 * Note:            None
 ********************************************************************/
static void HTTPLoadConn(BYTE connID)
{
    WORD oldPtr;
    
    // Return if already loaded
    if(connID == curHTTPID)
    	return;
    
    // Save the old one
    oldPtr = MACSetWritePtr(BASE_HTTPB_ADDR + curHTTPID*sizeof(HTTP_CONN));
	MACPutArray((BYTE*)&curHTTP, sizeof(HTTP_CONN));
	MACSetWritePtr(oldPtr);
	
	// Load the new one
    oldPtr = MACSetReadPtr(BASE_HTTPB_ADDR + connID*sizeof(HTTP_CONN));
	MACGetArray((BYTE*)&curHTTP, sizeof(HTTP_CONN));
	MACSetReadPtr(oldPtr);
	
	// Remember which one is loaded
	curHTTPID = connID;
}

// Perform the authentication checks. 
// This function will be called twice. First to determine if authentication
// will be necessary and a second time to check username/password.
BYTE HTTPAuthenticate(BOOL authCheck, BYTE *user, BYTE *pass) {

	if(authCheck) {
		// First round, determine if authentication is necessary
		 return (AppConfig.Flags.bUseAuthentication ? 0x00 : 0x80);

	}	else	{

		// Second round, check username/password combination
		if ((strcmp((char *) user, AppConfig.Username) == 0) && (strcmp((char *) pass, AppConfig.Password) == 0)) {
			return 0x80;		// We accept this combination
		}	else	{
			return 0x00;		// Provided user/pass is invalid
		}
	}
}

/*********************************************************************
 * Function:        static void HTTPProcess(void)
 *
 * PreCondition:    HTTPInit() called and curHTTP loaded
 *
 * Input:           None
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:		Serves the current HTTP connection in curHTTP
 *
 * Note:            None
 ********************************************************************/
static void HTTPProcess(void) {
    
	WORD lenA, lenB;
	BYTE c, i;
    BOOL b, done;
	BYTE *ext;
	BYTE buffer[HTTP_MAX_HEADER_LEN+1];
	char str[8];

    do {
		// Return will be made to the application code after each state
		// transition unless done is set FALSE within the state processing code.
		done = TRUE;

        // If a socket is disconnected during processing,
        // forget about it and return to idle state.
        if(TCPWasReset(sktHTTP)) {

			// Clean up file access
			if (curHTTP.handle != FFS_NO_HANDLE) {
				FFS_CleanUp(&curHTTP.handle);
			}

			// Adjust state machine state
            smHTTP = SM_HTTP_IDLE;

			// Adjust the TCP FIFOs for optimal reception of 
			// the next HTTP request from the browser
			TCPAdjustFIFOSize(sktHTTP, 1, 0, TCP_ADJUST_GIVE_REST_TO_RX | TCP_ADJUST_PRESERVE_RX);
         }

        switch(smHTTP) {

        case SM_HTTP_IDLE:

			// Check how much data is waiting
			lenA = TCPIsGetReady(sktHTTP);

			// If a connection has been made, then process the request
            if(lenA) {
            	// Clear out state info and move to next state
				curHTTP.ticcount = TickGet() + HTTP_TIMEOUT*TICK_SECOND;
				curHTTP.ptrData = curHTTP.data;
				curHTTP.fn[0] = '\0';
				curHTTP.fileExists = FALSE;
				curHTTP.handle = FFS_NO_HANDLE;
				curHTTP.size = 0;
				curHTTP.offset = 0;
				curHTTP.hasArgs = FALSE;
				curHTTP.isAuthorized = 0xff;
				curHTTP.httpStatus = HTTP_GET;
				curHTTP.fileType = HTTP_UNKNOWN;
				curHTTP.data[0] = '\0';

				smHTTP = SM_HTTP_PARSE_REQUEST;
			}
			break;

		case SM_HTTP_PARSE_REQUEST:

			// Verify the entire first line is in the FIFO
			if(TCPFind(sktHTTP, '\n', 0, FALSE) == 0xffff)
			{// First line isn't here yet
				if(TCPGetRxFIFOFree(sktHTTP) == 0)
				{// If the FIFO is full, we overflowed
					curHTTP.httpStatus = HTTP_OVERFLOW;
					smHTTP = SM_HTTP_SERVE_HEADERS;
					done = FALSE;
				}
				if(TickGet() > curHTTP.ticcount)
				{// A timeout has occurred
					TCPDisconnect(sktHTTP);
					smHTTP = SM_HTTP_DISCONNECT;
					done = FALSE;
				}
				break;
			}

			// Reset the watchdog timer
			curHTTP.ticcount = TickGet() + HTTP_TIMEOUT*TICK_SECOND;

			// Determine the request method
			lenA = TCPFind(sktHTTP, ' ', 0, FALSE);
			if(lenA > 5)
				lenA = 5;
			TCPGetArray(sktHTTP, curHTTP.data, lenA+1);

		    if ( memcmppgm2ram((void *) curHTTP.data, (const far rom void *) "GET", 3) == 0) {
			    curHTTP.httpStatus = HTTP_GET;
			}	else	{
				// Unrecognized method, so return not implemented
		        curHTTP.httpStatus = HTTP_NOT_IMPLEMENTED;
				smHTTP = SM_HTTP_SERVE_HEADERS;
				done = FALSE;
				break;
			}

			// Find end of filename
			lenA = TCPFind(sktHTTP, ' ', 0, FALSE);
			lenB = TCPFindEx(sktHTTP, '?', 0, lenA, FALSE);
			lenA = mMIN(lenA, lenB);
			
			// If the file name is too long, then reject the request
			if(lenA > HTTP_MAX_DATA_LEN - HTTP_DEFAULT_LEN - 1) {
				curHTTP.httpStatus = HTTP_OVERFLOW;
				smHTTP = SM_HTTP_SERVE_HEADERS;
				done = FALSE;
				break;
			}

			// Read in the filename
			lenB = TCPGetArray(sktHTTP, curHTTP.data, lenA);
			curHTTP.data[lenB] = '\0';
			
			// Is the last character a directory delimiter ?
			if(curHTTP.data[lenB-1] == '/') {
				// It is, so add the default file
				strcpypgm2ram((char *)&curHTTP.data[lenB], (char *) HTTP_DEFAULT_FILE);
				lenB += HTTP_DEFAULT_LEN;
			}
			// Determine if the specified file exists
			curHTTP.fileExists = FFS_FileExists((char *) curHTTP.data, &curHTTP.size);
			if (curHTTP.fileExists) {
				strcpy(curHTTP.fn, (char * ) curHTTP.data);
			}
			
			// Find the extension in the filename
			for(ext = curHTTP.data + lenB-1; ext != curHTTP.data; ext--)
				if(*ext == '.')
					break;
					
			// Compare to known extensions to determine Content-Type
			ext++;
			for(curHTTP.fileType = HTTP_TXT; curHTTP.fileType < HTTP_UNKNOWN; curHTTP.fileType++)
				if(!stricmppgm2ram(ext, (ROM void*)httpFileExtensions[curHTTP.fileType]))
					break;
			
			// Perform first round authentication (pass file name only)
			curHTTP.isAuthorized = HTTPAuthenticate(TRUE, NULL, NULL);
			
			// Read GET args, up to buffer size - 1
			lenA = TCPFind(sktHTTP, ' ', 0, FALSE);
			if(lenA != 0)
			{
				curHTTP.hasArgs = TRUE;
				
				// Trash the '?'
				TCPGet(sktHTTP, &c);

				// Verify there's enough space
				lenA--;
				if(lenA >= HTTP_MAX_DATA_LEN - 2)
				{
			        curHTTP.httpStatus = HTTP_OVERFLOW;
					smHTTP = SM_HTTP_SERVE_HEADERS;
					done = FALSE;
					break;
				}
				// Read in the arguments and '&'-terminate in anticipation of cookies
				curHTTP.ptrData += TCPGetArray(sktHTTP, curHTTP.data, lenA);
				*(curHTTP.ptrData++) = '&';
			}

			// Clear the rest of the line
			lenA = TCPFind(sktHTTP, '\n', 0, FALSE);
			TCPGetArray(sktHTTP, NULL, lenA + 1);

			// Move to parsing the headers
			smHTTP = SM_HTTP_PARSE_HEADERS;
			
			// No break, continue to parsing headers
			
		case SM_HTTP_PARSE_HEADERS:

			// Loop over all the headers
			while(1)
			{
				// Make sure entire line is in the FIFO
				lenA = TCPFind(sktHTTP, '\n', 0, FALSE);
				if(lenA == 0xffff)
				{// If not, make sure we can receive more data
					if(TCPGetRxFIFOFree(sktHTTP) == 0)
					{// Overflow
						curHTTP.httpStatus = HTTP_OVERFLOW;
						smHTTP = SM_HTTP_SERVE_HEADERS;
						done = FALSE;
					}
					if(TickGet() > curHTTP.ticcount)
					{// A timeout has occured
						TCPDisconnect(sktHTTP);
						smHTTP = SM_HTTP_DISCONNECT;
						done = FALSE;
					}
					break;
				}
				
				// Reset the watchdog timer
				curHTTP.ticcount = TickGet() + HTTP_TIMEOUT*TICK_SECOND;
				
				// If a CRLF is immediate, then headers are done
				if(lenA == 1) {
					// Remove the CRLF and move to next state
					TCPGetArray(sktHTTP, NULL, 2);
					smHTTP = SM_HTTP_AUTHENTICATE;
					done = FALSE;
					break;
				}
	
				// Find the header name, and use done as a flag to indicate a match
				lenB = TCPFindEx(sktHTTP, ':', 0, lenA, FALSE) + 2;
				done = FALSE;
	
				// If name is too long or this line isn't a header, ignore it
				if(lenB > sizeof(buffer)) {
					TCPGetArray(sktHTTP, NULL, lenA+1);
					continue;
				}
				
				// Read in the header name
				TCPGetArray(sktHTTP, buffer, lenB);
				buffer[lenB-1] = '\0';
				lenA -= lenB;
		
				// Compare header read to ones we're interested in
				for(i = 0; i < HTTP_NUM_HEADERS; i++) {
					if(strcmppgm2ram((char*)buffer, (const rom far char *) HTTPRequestHeaders[i]) == 0) {
						// Parse the header and stop the loop
						HTTPHeaderParseLookup(i);
						done = TRUE;
						break;
					}
				}
				
				// Clear the rest of the line, and call the loop again
				if(done)
				{// We already know how much to remove unless a header was found
					lenA = TCPFind(sktHTTP, '\n', 0, FALSE);
				}
				TCPGetArray(sktHTTP, NULL, lenA+1);
			}
			break;

		case SM_HTTP_AUTHENTICATE:
		
			// Check current authorization state
			if(curHTTP.isAuthorized < 0x80) {
				// 401 error
				curHTTP.httpStatus = HTTP_UNAUTHORIZED;
				smHTTP = SM_HTTP_SERVE_HEADERS;
				done = FALSE;
				break;
			}

			// Move to next state
			smHTTP = SM_HTTP_PROCESS_REQUEST;
			break;

		case SM_HTTP_PROCESS_REQUEST:

			// Check for 404
            if(!curHTTP.fileExists) {
                curHTTP.httpStatus = HTTP_NOT_FOUND;
                smHTTP = SM_HTTP_SERVE_HEADERS;
				done = FALSE;
                break;
            }

			// Move to next state
			smHTTP = SM_HTTP_SERVE_HEADERS;
			break;

		case SM_HTTP_SERVE_HEADERS:

			// We're in write mode now:
			// Adjust the TCP FIFOs for optimal transmission of 
			// the HTTP response to the browser
			TCPAdjustFIFOSize(sktHTTP, 1, 0, TCP_ADJUST_GIVE_REST_TO_TX);
				
			// Send headers
			TCPPutROMString(sktHTTP, (ROM BYTE*)HTTPResponseHeaders[curHTTP.httpStatus]);
			
			// If this is a redirect, print the rest of the Location: header			   
			if(curHTTP.httpStatus == HTTP_REDIRECT)
			{
				TCPPutString(sktHTTP, curHTTP.data);
				TCPPutROMString(sktHTTP, (ROM BYTE*)"\r\n\r\n304 Redirect: ");
				TCPPutString(sktHTTP, curHTTP.data);
				TCPPutROMString(sktHTTP, (ROM BYTE*)HTTP_CRLF);
			}

			// If not GET, we're done
			if (curHTTP.httpStatus != HTTP_GET) {
				// Disconnect
				smHTTP = SM_HTTP_DISCONNECT;
				break;
			}

			// Output the content type, if known
			if(curHTTP.fileType != HTTP_UNKNOWN)
			{
				TCPPutROMString(sktHTTP, (ROM BYTE*)"Content-Type: ");
				TCPPutROMString(sktHTTP, (ROM BYTE*)httpContentTypes[curHTTP.fileType]);
				TCPPutROMString(sktHTTP, HTTP_CRLF);
			}
			
			// Output the content length
			TCPPutROMString(sktHTTP, (ROM BYTE*)"Content-Length: ");
			str[0] = '\0';
			ltoa(curHTTP.size, str);
			TCPPutString(sktHTTP, (BYTE *) str);
			TCPPutROMString(sktHTTP, HTTP_CRLF);

			// Output the cache-control
			TCPPutROMString(sktHTTP, (ROM BYTE*)"Cache-Control: ");
			TCPPutROMString(sktHTTP, (ROM BYTE*)"max-age=");
			TCPPutROMString(sktHTTP, (ROM BYTE*)HTTP_CACHE_LEN);
			TCPPutROMString(sktHTTP, HTTP_CRLF);

			// Terminate the headers
			TCPPutROMString(sktHTTP, HTTP_CRLF);

			// Move to next state
			smHTTP = SM_HTTP_SERVE_BODY;
			break;

		case SM_HTTP_SERVE_BODY:

//			done = FALSE;

			// Try to send next packet
			if (HTTPSendFile()) {
				// If EOF, we're done so disconnect
				smHTTP = SM_HTTP_DISCONNECT;
//				done = TRUE;
			}
			break;

		case SM_HTTP_DISCONNECT:
		
			// Loopbacks have no wait state, so all data must be retrieved first
			if(TCPIsLoopback(sktHTTP) && TCPGetTxFIFOFull(sktHTTP) != 0)
				break;

			curHTTP.handle = FFS_NO_HANDLE;
			TCPDisconnect(sktHTTP);
            smHTTP = SM_HTTP_IDLE;
            break;
		}
	} while (!done);
}

/*********************************************************************
 * Function:        static BOOL HTTPSendFile(void)
 *
 * PreCondition:    File to send is known to exist
 *
 * Input:           None
 *
 * Output:          TRUE if EOF was reached and reading is done
 *					FALSE if more data remains
 *
 * Side Effects:    None
 *
 * Overview:        This function serves the next chunk of file data
 *
 * Note:            None
 ********************************************************************/
static BOOL HTTPSendFile(void)
{
	WORD numBytes;
	BYTE c, *buffer;
	DWORD readsize;
	
	// Prepare for reading the file
	c = FFS_GetFileHandle((char * ) curHTTP.fn, &curHTTP.handle);
	if ((c != FFS_NO_ERROR) || (curHTTP.handle == FFS_NO_HANDLE)) {
		return FALSE;
	}

	// Determine how many bytes we can send right now
	numBytes = mMIN(TCPIsPutReady(sktHTTP), FILE_READ_SIZE);
	if (numBytes <= 0) {
		// If we cannot send any data, just return to let
		// transmit buffers empty.
		return FALSE;
	}

	// Every handle has a dedicated data buffer
	switch (curHTTP.handle) {

		case 0:
			buffer = data0;
			break;
		case 1:
			buffer = data1;
			break;
		case 2:
			buffer = data2;
			break;
		case 3:
			buffer = data3;
			break;
	}
	
	// Attempt to read numBytes of data
	c = FFS_ReadFile(&curHTTP.handle, buffer, &curHTTP.offset, numBytes, &readsize);
	if ((c == FFS_EOF) || (c != FFS_NO_ERROR)) {
		// If we have reached EOF or a read error has occurred, terminate file read
		return TRUE;

	}	else	{
		// We have good data so send it out
		TCPPutArray(sktHTTP, buffer, readsize);
	}
    // We are not done sending a file yet...
    return FALSE;
}

/*********************************************************************
 * Function:        static void HTTPHeaderParseLookup(BYTE i)
 *
 * PreCondition:    None
 *
 * Input:           i: the index of the header parser found
 *
 * Output:          None
 *
 * Side Effects:    None
 *
 * Overview:        Calls the appropriate header parser
 *
 * Note:            None
 ********************************************************************/
static void HTTPHeaderParseLookup(BYTE i)
{
	// i corresponds to an index in HTTPRequestHeaders
	
	if (i == 0u)
	{
		HTTPHeaderParseAuthorization();
		return;
	}
}

/*********************************************************************
 * Function:        static void HTTPHeaderParseAuthorization(void)
 *
 * PreCondition:    None
 *
 * Input:           None
 *
 * Output:          myConn:	isAuthorized is updated
 * 
 * Side Effects:    None
 *
 * Overview:        Parses the "Authorization:" header.
 *					For example, "BASIC YWRtaW46cGFzc3dvcmQ=\r\n" is:
 *						admin(\0)password(\0)
 *						^        ^
 *					    buf      ptrBuf
 *					Once read, the user's authorization function is 
 *					called to process and store a result.
 *
 * Note:            Tries to read up to and including the end of the 
 *					current line.  If not successful, a return of
 *					FALSE indicates that this function should be 
 *					called again when more data is available.
 ********************************************************************/
static void HTTPHeaderParseAuthorization(void)
{
    WORD len;
    BYTE buf[40];
	BYTE *ptrBuf;
	
	// If auth processing is not required, return
	if(curHTTP.isAuthorized & 0x80)
		return;

	// Clear the auth type ("BASIC ")
	TCPGetArray(sktHTTP, NULL, 6);

	// Find the terminating CRLF and make sure it's a multiple of four
	len = TCPFindROMArray(sktHTTP, HTTP_CRLF, HTTP_CRLF_LEN, 0, FALSE);
	len += 3;
	len &= 0xfc;
	len = mMIN(len, sizeof(buf)-4);
	
	// Read in 4 bytes at a time and decode (slower, but saves RAM)
	for(ptrBuf = buf; len > 0; len-=4, ptrBuf+=3)
	{
		TCPGetArray(sktHTTP, ptrBuf, 4);
		Base64Decode(ptrBuf, 4, ptrBuf, 3);
	}

	// Null terminate both, and make sure there's at least two terminators
	*ptrBuf = '\0';
	for(len = 0, ptrBuf = buf; len < sizeof(buf); len++, ptrBuf++)
		if(*ptrBuf == ':')
			break;
	*(ptrBuf++) = '\0';
	
	// Verify credentials
	curHTTP.isAuthorized = HTTPAuthenticate(FALSE, buf, ptrBuf);

	return;
}


