/*********************************************************************
 *  Main Entry Point for Webserver
 *  Using Microchip TCP/IP Stack
 *
 * Author: Craig A. Lindley
 * Last Update: 11/22/2007
 ********************************************************************/

#include "ctype.h"
#include "TCPIP Stack/TCPIP.h"
#include "TCPIP Stack/HTTP2.h"
#include "TCPIP Stack/Ffs.h"

// Set configuration
#pragma config OSC=HS, WDT=OFF, MCLRE=ON, PBADEN=OFF, LVP=OFF

// Delay count for error display
#define DELAY_COUNT 20000

// Define the configuration file's name
#define CONFIG_FILE_NAME "config.dat"
#define MAX_CONFIG_SIZE 500
#define MAX_CONFIG_LINE 60

// Buffer to use temporarily for reading in the configuration data
extern BYTE data0[];
BOOL configFileRead = FALSE;

// Define and initialize the app config structure which is used by
// various stack elements. The content of this structure
// will be read from the configuration file.
APP_CONFIG AppConfig;

//
// PIC18 Interrupt Service Routines
// 
// NOTE: Several PICs, including the PIC18F4620 revision A3 have a RETFIE FAST/MOVFF bug
// The interruptlow keyword is used to work around the bug when using C18
#pragma interruptlow LowISR
void LowISR(void) {
    TickUpdate();
}
#pragma code lowVector=0x18
void LowVector(void){_asm goto LowISR _endasm}
#pragma code

// Initialize the board specific hardware
void InitializeBoard(void) {

	// UART
	UARTTX_TRIS = 0;
	UARTRX_TRIS = 1;

	// Set LED I/O lines to outputs
	LED0_TRIS = 0;
	LED1_TRIS = 0;
	LED2_TRIS = 0;

	// Turn the LEDs off
	LED0_IO = 0;
	LED1_IO = 0;
	LED2_IO = 0;

	// Encoder I/O lines
	ENC_RST_TRIS = 0;
	ENC_CS_TRIS = 0;
	ENC_SCK_TRIS = 0;
	ENC_SDI_TRIS = 1;
	ENC_SDO_TRIS = 0;

	ENC_RST_IO = 0;
	ENC_CS_IO = 1;

    // Enable internal PORTB pull-ups
    INTCON2bits.RBPU = 0;

	// Enable Interrupts
	RCONbits.IPEN = 1;		// Enable interrupt priorities
    INTCONbits.GIEH = 1;
    INTCONbits.GIEL = 1;
}

// NOTE: Name[] must be at least 16 characters long.
// It should be exactly 16 characters, as defined by NetBIOS spec.
void FormatNetBIOSName(BYTE Name[]) {
	BYTE i;

	Name[15] = '\0';
	strupr((char*)Name);
	i = 0;
	while(i < 15u) {
		if(Name[i] == '\0') {
			while(i < 15u) {
				Name[i++] = ' ';
			}
			break;
		}
		i++;
	}
}

// Advance pointer to past colon and space
char * adv(char *line) {

	char c, *ptr = line;
	c = *ptr++;
	while (c != ':') {
		c = *ptr++;
	}
	return ++ptr;
}

// Ascii hex to integer conversion
BYTE htoi(char *buffer) {

	BYTE l;
	int msb, lsb;

	msb = 0;
	l = strlen(buffer);
	if (l > 1) {
		msb = buffer[0];
		msb = tolower(msb);
		msb -= isdigit(msb) ? 0x30 : 0x57;
		lsb = buffer[1];
	}	else	{
		lsb = buffer[0];
	}
	lsb = tolower(lsb);
	lsb -= isdigit(lsb) ? 0x30 : 0x57;

	return ((BYTE) lsb | msb<<4u);
}

// Parse a MAC address of the form 00-04-A3-00-47-F7
void parseMAC(char *line, BYTE macAddress[]) {

	BYTE i, j, MACAddress[6];
	char c, *ptr, buffer[3];

	// Advance to the start of the MAC address
 	ptr = adv(line);
	for (i = 0; i < 6; i++) {
		c = *ptr++;
		j = 0;
		while ((c != '-') && (c != '\0')) {
			buffer[j++] = c;
			c = *ptr++;
		}
		buffer[j] = '\0';
		macAddress[i] = htoi(buffer);
	}
}


// Parse an IP address into a DWORD
DWORD parseIP(char *ipStr) {

	BYTE i, j, bytes[4];
	DWORD IPAddress;
	char c, buffer[4];

	ipStr = adv(ipStr);

	for (i = 0; i < 4; i++) {
		c = *ipStr++;
		j = 0;
		while ((c != '.') && (c != '\0')) {
			buffer[j++] = c;
			c = *ipStr++;
		}
		buffer[j] = '\0';
		bytes[i] = atoi(buffer);
	}
	return ((DWORD) bytes[0] | ((DWORD) bytes[1])<<8ul | ((DWORD) bytes[2])<<16ul | ((DWORD) bytes[3])<<24ul);
}

// Short delay function
void delay(BYTE loops) {
	WORD loop, count;

	if (loops <=0) {
		loops = 1;
	}
	for (loop = 0; loop < loops; loop++) {
		for (count = 0; count < DELAY_COUNT; count++) {
			;
		}
	}
}

// Signal configuration error by flashing all onboard LEDs forever.
// These is no return from the subroutine.
void signalError(void) {

	// Do until user resets the hardware
	while (1) {

		// Pre delay
		delay(1);

		// Toggle the LEDs
		LED0_IO ^= 1;
		LED1_IO ^= 1;
		LED2_IO ^= 1;

		// Post delay
		delay(1);
	}
}

// Read in the configuration data and return a line each call
BYTE readLine(char *buffer) {

	BOOL b;
	BYTE r, handle;
	static BYTE *ptr;
	char c, fn[12];
	DWORD size;

	DWORD offset, readsize;

	// Has the configuration file already been read in ?
	if (!configFileRead) {
		// No, so attempt to read it in now
		offset = 0;
		handle = FFS_NO_HANDLE;

		strcpypgm2ram(fn, (const far rom char *) CONFIG_FILE_NAME);

		// Does the configuation file exist ?
		b = FFS_FileExists(fn, &size);
		if (b != TRUE) {
			// Configuration file not found
			signalError();
		}
		// Attempt to get a handle to the file for reading
		r = FFS_GetFileHandle(fn, &handle);
		if (r != FFS_NO_ERROR) {
			// Couldn't open the file so signal error
			signalError();
		}
		// File is opened for reading
		r = FFS_ReadFile(&handle, data0, &offset, MAX_CONFIG_SIZE, &readsize);
		if (r != FFS_NO_ERROR) {
			// Couldn't read file so signal error
			signalError();
		}
		FFS_CleanUp(&handle);

		// Indicate file has been read into memory
		configFileRead = TRUE;

		// Initialize pointer to the data in memory
		ptr = (BYTE *) data0;
	}
	// Return a line from the configuration file
	c = *ptr++;
	if ((c == '\r') || (c == '\0')) {
		return FFS_EOF;
	}

	while (c != '\r') {
		*buffer++ = c;
		c = *ptr++;
	}
	*buffer = '\0';
	ptr++;

	return FFS_NO_ERROR;
}

// Read in the configuration file and set the server
// to this configuration. Filesystem must be initialized
// before this call is made.
void InitAppConfig(void) {

	char line[MAX_CONFIG_LINE], *ptr;
	BYTE c;

	// Clear the AppConfig structure to begin with
	memset((void *) &AppConfig, '\0', sizeof(APP_CONFIG));

	// Initialize known members
	AppConfig.Flags.bInConfigMode = FALSE;

	while (1) {
		// Read a line from the configuration file
		c = readLine(line);
		if (c == FFS_EOF) {
			break;
		}
	    if (memcmppgm2ram((void *) line, (const far rom void *) "//", 2) == 0) {
			continue;

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Host", 4) == 0) {
			ptr = adv(line);
			strcpy((char * ) AppConfig.NetBIOSName, ptr);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Use D", 5) == 0) {
			ptr = adv(line);
			AppConfig.Flags.bIsDHCPEnabled = (*ptr == 'y') ? TRUE : FALSE;

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Use A", 5) == 0) {
			ptr = adv(line);
			AppConfig.Flags.bUseAuthentication = (*ptr == 'y') ? TRUE : FALSE;

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "User", 4) == 0) {
			ptr = adv(line);
			strcpy(AppConfig.Username, ptr);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Pass", 4) == 0) {
			ptr = adv(line);
			strcpy(AppConfig.Password, ptr);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "MAC", 3) == 0) {
			parseMAC(line, AppConfig.MyMACAddr.v);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Static I", 8) == 0) {
			AppConfig.MyIPAddr.Val = parseIP(line);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Static G", 8) == 0) {
			AppConfig.MyGateway.Val = parseIP(line);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Static Su", 9) == 0) {
			AppConfig.MyMask.Val = parseIP(line);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Static P", 8) == 0) {
			AppConfig.PrimaryDNSServer.Val = parseIP(line);

	    }	else if (memcmppgm2ram((void *) line, (const far rom void *) "Static Se", 9) == 0) {
			AppConfig.SecondaryDNSServer.Val = parseIP(line);
		}
	}
	AppConfig.DefaultIPAddr.Val = AppConfig.MyIPAddr.Val;
	AppConfig.DefaultMask.Val = AppConfig.MyMask.Val;
	FormatNetBIOSName(AppConfig.NetBIOSName);
}

//
// Main application entry point.
//
void main(void) {

    static TICK t = 0;

    // Initialize hardware.
    InitializeBoard();

    // Initialize FAT file system hardware
    FFS_Init(USB_DEVICE_PORT_1);

	// Do a short delay
	delay(20);

    // Initialize all stack related components.
    // Following steps must be performed for all applications using
    // the Microchip TCP/IP Stack.
    TickInit();

     // Initialize stack from configuration file.
    InitAppConfig();

	// Initialize core stack layers (MAC, ARP, TCP, UDP)
    StackInit();

	// Initialize all HTTP connections for running
    HTTPInit();

    if (!AppConfig.Flags.bIsDHCPEnabled) {
        DHCPDisable();
    }

    // Once all items are initialized, go into infinite loop and let
    // stack items execute their tasks.
    // If application needs to perform its own task, it should be
    // done at the end of while loop.
    // Note that this is a "co-operative mult-tasking" mechanism
    // where every task performs its tasks (whether all in one shot
    // or part of it) and returns so that other tasks can do their
    // job.
    // If a task needs very long time to do its job, it must be broken
    // down into smaller pieces so that other tasks can have CPU time.
    while(1) {

        // Blink LED0 every second.
        if (TickGet() - t >= TICK_SECOND/2ul) {
            t = TickGet();
            LED0_IO ^= 1;
        }
        // This task performs normal stack task including checking
        // for incoming packet, type of packet and calling
        // appropriate stack entity to process it.
        StackTask();

        // This is a TCP application.  It listens to TCP port 80
        // with one or more sockets and responds to remote requests.
        HTTPServer();
    }
}
